;This DSP program implements a basic hard limiter compressor.
;If the signal clips, the compressor reduces the gain to preserve the shape of the signal.
;This program accepts 1-2 user arguments:
;float - maximum amplitude, non-negative, default 1.0. Caps the signal to this amplitude.
;float - linear gain recovery rate, per second, non-negative, default 1.0. 
;For instance, if the recovery rate is 0.5, the gain factor takes up to 2 seconds to recover.

jmp .initializer        ;jump over data segment

$sampleRate dword 0.0   ;Sample rate, set by initializer
$channels   dword 0     ;Channel count, set by initializer
$gain       dword 1.0   ;Current gain multiplier
$maxAmp     dword 1.0   ;Maximum output amplitude
$recoveryHz dword 1.0   ;Gain recovery rate per second
$recovery   dword 0.0   ;Gain recovery rate per frame

.initializer:
    lcons t0, 4         ;load constant 4
    add sp, sp, t0      ;ca_dsp version (don't care)
    pop t1              ;sample rate
    i2f t1, t1
    stor $sampleRate, t1
    pop t2              ;channel count
    stor $channels, t2
    
    ;Compute recovery rate per frame
    lcons t4, $recoveryHz
    ;Divide recovery rate by sample rate
    fdiv t4, t4, t1
    stor $recovery, t4
    
    lcons r0, .msgProc
    lcons r1, .msgProc
    lcons r2, .main
    halt
    
.msgProc:
    pop t3          ;user argument count
    jz t3, .msgProcEnd
        ;Has maxAmp argument
        pop t1
        stor $maxAmp, t1
        dec t3
    jz t3, .msgProcEnd
        ;Has recovery argument
        pop r3
        stor $recoveryHz, r3
        load t1, $sampleRate
        fdiv r3, r3, t1
        stor $recovery, r3
    .msgProcEnd:
    
    lcons r0, 0
    halt

.main:
	pop r2              ;buffer size in frames
	pop r1              ;buffer address
    
    load r3, $recovery
    load r0, $gain
	load t0, $channels
	
	mul r2, t0, r2     ;buffer size = channels*frames
    
    ;Summary
    ;r0 gain, r1 buffer address, r2 sample count, r3 recovery rate
	
    lcons t0, 0        ;constant 0
    load t1, $maxAmp
    lcons t2, -1.0
    fmul t2, t2, t1     ;negative maxAmp
	lcons t3, 4        ;constant 4 bytes per sample
    
	;Loop over each input sample, amplify it and place it in the destination buffer 
	lcons t4, 0        ;index
	
	.loopStart:
		;Terminate when index reaches sample count
		jae t4, r2, .loopEnd
        
        je r0, t1, .gainFine
            fadd r0, r0, r3 ;increment gain
            fcom t5, r0, t1 ;check if gain > maxAmp
            jle t5, t0, .gainFine
                lcons t8, 87
                mov r0, t1       ;set gain to maxAmp if greater
        .gainFine:
            
		load_p t8, r1   ;fetch input sample
		fmul t6, t8, r0 ;amplify, sample *= gain
        
        fcom t5, t6, t1 ;check if sample > maxAmp
        jle t5, t0, .sampleLe1
            ;Amplified sample is greater than maxAmp
            ;Update gain to maxAmp/originalsample
            fdiv r0, t1, t8
            ;clamp the sample value to maxAmp
            mov t6, t1
            jmp .sampleFine
        .sampleLe1:
            fcom t5, t6, t2 ;check if sample < -maxAmp
            jge t5, t0, .sampleFine
                ;Amplified sample is less than -maxAmp
                ;Update gain to -maxAmp/originalsample
                fdiv r0, t2, t8
	            ;clamp the sample value to -maxAmp
	            mov t6, t2
        .sampleFine:
        
		stor_p r1, t6   ;store output sample
		
		;Increment index, buffer address
		inc t4
		add r1, r1, t3
		jmp .loopStart
	.loopEnd:
	
    
	;Store current gain and return 0 for success
    stor $gain, r0
	lcons r0, 0
	halt