;Competent Audio comes with a variant of Jezar's Freeverb built in (in native code).
;That functionality is exposed to DSPs via interrupts [0x70-0x78] (see ca_dsp_interrupts).
;This program runs the reverberator on the audio data and allows the application to 
;change the reverb parameters via user argument commands. 
;
;Each command consists of one or two arguments, where the first is an integer ID,
;and the second is a floating point value:
;ID  Name         Value
;0   Mute         N/A
;    Muting resets the reverberator's state. This can be used for sharp transitions.
;1   Room size    Room size factor
;    A higher room size increases the intensity of the reverb signal.
;2   Damp         Damping factor
;    A lower damping factor causes the reverb signal to diminish faster.
;3   Wet          Wet signal factor
;    The wet factor multiplies the reverberation component of the signal.
;4   Dry          Dry signal factor
;    The dry factor multiplies the original unmodified component of the signal.
;5   Width        Stereo width factor
;    A higher stereo width increases the feedback between left and right channels.
;6   Freeze mode  1.0 to enable, 0.0 to disable
;    Freeze mode sustains the current wet signal and prevents inputs from affecting it.
;
;The mute command expects no value, so there is no value argument.
;All other commands require a value argument.
;Multiple commands can be specified in sequence, for instance:
;    [0, 3, 0.5, 4, 0.75]
;mutes the reverb, sets the wet factor to 0.5 and the dry factor to 0.75.

jmp .initializer

$reverb     dword 0
$rate       dword 0
$channels   dword 0
$roomSize	dword 0.5
$damp		dword 0.5
$wet		dword 0.33
$dry		dword 0.33
$width	    dword 1.0
$freezeMode	dword 0.0

.initializer:
    lcons t4, 4         ;constant 4 bytes
    
    add sp, sp, t4      ;skip version number
    pop r0              ;sample rate
    stor $rate, r0      ;save sample rate
    pop r1              ;channel count
    stor $channels, r1  ;save channel count
    
    lcons r0, .initMP   ;return init message processor
    lcons r1, .mainMP   ;return main message processor
    lcons r2, .main     ;return main routine
    halt

.initMP:
    lcons t0, 0         ;set reverb handle to null,
    stor $reverb, t0    ;in case of re-initialization
    
    mov t0, sp          ;pointer to user argument list
    int 0               ;enter
    push t0             ;pass user argument list
    call .args          ;process user arguments
    int 1               ;exit
    
    
    load r0, $rate
    load r1, $channels
    
    int 0               ;enter
    load t0, $freezeMode
    push t0             ;arg7
    load t0, $width
    push t0             ;arg6
    load t0, $dry
    push t0             ;arg5
    load t0, $wet
    push t0             ;arg4
    load t0, $damp
    push t0             ;arg3
    load t0, $roomSize
    push t0             ;arg2
    push r1             ;arg1 - channel count
    push r0             ;arg0 - sample rate
    int 0x70            ;create reverb
    stor $reverb, r0    ;save handle
    int 1               ;exit
    
    lcons r0, 0
    halt

.mainMP:
    mov t0, sp          ;pointer to user argument list
    int 0               ;enter
    push t0             ;pass user argument list
    call .args          ;process user arguments
    int 1               ;exit
    
    lcons r0, 0
    halt

.main:
    pop r2              ;r2 buffer size in frames
    pop r1              ;r1 buffer address
	
    load t0, $reverb
	load t3, $channels
    int 0              ;enter
    push t3            ;arg 3 - frame skip
    push r2            ;arg 2 - buffer size
    push r1            ;arg 1 - buffer address
    push t0            ;arg 0 - reverb handle
    int 0x71            ;process reverb
    int 1              ;exit

	lcons r0, 0        ;return success
	halt

.args:
    lcons t4, 4         ;constant 4 bytes
    
    pop t5              ;user argument list pointer
    load_p t0, t5       ;list size
    load t1, $reverb    ;reverberator handle
    
    .argsStart:
        jz t0, .argsEnd     ;return when no more args
        
        add t5, t5, t4      ;next list item
        load_p t2, t5       ;command id
        dec t0;             ;decrease remaining arg count
        
        lcons t6, 0         ;command 0: mute, takes no operand
        je t2, t6, .doMute
        
        jnz t0, .argOperand
            lcons r0, -6        ;user error: missing operand
            halt
        .argOperand:
        
        add t5, t5, t4      ;next list item
        load_p t3, t5       ;operand
        dec t0;             ;decrease remaining arg count
        
        lcons t6, 1         ;command 1: room size
        je t2, t6, .doRoomSize
        lcons t6, 2         ;command 2: damp
        je t2, t6, .doDamp
        lcons t6, 3         ;command 3: wet
        je t2, t6, .doWet
        lcons t6, 4         ;command 4: dry
        je t2, t6, .doDry
        lcons t6, 5         ;command 5: width
        je t2, t6, .doWidth
        lcons t6, 6         ;command 6: freeze mode
        je t2, t6, .doFreezeMode
        
        lcons r0, -92       ;user error: command id out of bounds
        halt
        
        
        .doMute:              ;command 0
            ;This seems pointless but it can happen on reinit
            jz t1, .argsStart
            int 0               ;enter
            push t1             ;handle
            int 0x72             ;mute
            int 1               ;exit
            jmp .argsStart
        .doRoomSize:          ;command 1
            stor $roomSize, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x73             ;roomSize
            int 1               ;exit
            jmp .argsStart
        .doDamp:              ;command 2
            stor $damp, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x74             ;damp
            int 1               ;exit
            jmp .argsStart
        .doWet:               ;command 3
            stor $wet, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x75             ;wet
            int 1               ;exit
            jmp .argsStart
        .doDry:               ;command 4
            stor $dry, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x76             ;dry
            int 1               ;exit
            jmp .argsStart
        .doWidth:             ;command 5
            stor $width, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x77             ;width
            int 1               ;exit
            jmp .argsStart
        .doFreezeMode:        ;command 6
            stor $freezeMode, t3
            jz t1, .argsStart
            int 0               ;enter
            push t3             ;operand
            push t1             ;handle
            int 0x78             ;freezeMode
            int 1               ;exit
            jmp .argsStart
        
    .argsEnd:
    ret